import base64
import json
import os
import re
import time
from datetime import datetime, tzinfo, timedelta

import jwt
import responses
from cryptography import x509
from cryptography.hazmat.backends import default_backend as crypto_default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization as crypto_serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.x509.oid import NameOID


def generate_key_and_cert():
    signing_key = rsa.generate_private_key(
        backend=crypto_default_backend(),
        public_exponent=65537,
        key_size=2048
    )
    subject = issuer = x509.Name([
        x509.NameAttribute(NameOID.COUNTRY_NAME, u"US"),
        x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, u"CA"),
        x509.NameAttribute(NameOID.LOCALITY_NAME, u"San Francisco"),
        x509.NameAttribute(NameOID.ORGANIZATION_NAME, u"My Company"),
        x509.NameAttribute(NameOID.COMMON_NAME, u"example.com"),
    ])
    signing_cert = x509.CertificateBuilder().subject_name(
        subject
    ).issuer_name(
        issuer
    ).public_key(
        signing_key.public_key()
    ).serial_number(
        x509.random_serial_number()
    ).not_valid_before(
        datetime.utcnow()
    ).not_valid_after(
        # Our certificate will be valid for 10 days
        datetime.utcnow() + timedelta(days=10)
        # Sign our certificate with our private key
    ).sign(
        signing_key, hashes.SHA256(), crypto_default_backend()
    ).public_bytes(crypto_serialization.Encoding.DER)
    return signing_key, signing_cert


class SimpleUtc(tzinfo):
    def tzname(self, dt):
        return "UTC"

    def utcoffset(self, dt):
        return timedelta(0)


def load_json(file):
    with open(os.path.join(os.path.dirname(__file__), file), mode="r") as f:
        data = json.load(f)
    return data


def build_access_token_adfs(request):
    issuer = "http://adfs.example.com/adfs/services/trust"
    return do_build_access_token(request, issuer)


def build_access_token_azure(request):
    issuer = "https://sts.windows.net/01234567-89ab-cdef-0123-456789abcdef/"
    return do_build_access_token(request, issuer)


def do_build_access_token(request, issuer):
    issued_at = int(time.time())
    expires = issued_at + 3600
    auth_time = datetime.utcnow()
    auth_time = auth_time.replace(tzinfo=SimpleUtc(), microsecond=0)
    claims = {
        "aud": "microsoft:identityserver:your-RelyingPartyTrust-identifier",
        "iss": issuer,
        "iat": issued_at,
        "exp": expires,
        "winaccountname": "testuser",
        "group": ["group1", "group2"],
        "given_name": "John",
        "family_name": "Doe",
        "email": "john.doe@example.com",
        "sub": "john.doe@example.com",
        "user_is_staff": "True",
        "user_is_superuser": "yes",
        "appid": "your-configured-client-id",
        "auth_time": auth_time.isoformat(),
        "authmethod": "urn:oasis:names:tc:SAML:2.0:ac:classes:PasswordProtectedTransport",
        "ver": "1.0"
    }
    token = jwt.encode(claims, signing_key_B, algorithm="RS256")
    response = {
        'resource': 'django_website.adfs.relying_party_id',
        'token_type': 'bearer',
        'refresh_token_expires_in': 28799,
        'refresh_token': 'random_refresh_token',
        'expires_in': 3600,
        'id_token': 'not_used',
        'access_token': token.decode()
    }
    return 200, [], json.dumps(response)


def build_openid_keys(request):
    keys = {
        "keys": [
            {
                "kty": "RSA",
                "use": "sig",
                "kid": "dummythumbprint",
                "x5t": "dummythumbprint",
                "n": "somebase64encodedmodulus",
                "e": "somebase64encodedexponent",
                "x5c": [base64.b64encode(signing_cert_A).decode(), ]
            },
            {
                "kty": "RSA",
                "use": "sig",
                "kid": "dummythumbprint",
                "x5t": "dummythumbprint",
                "n": "somebase64encodedmodulus",
                "e": "somebase64encodedexponent",
                "x5c": [base64.b64encode(signing_cert_B).decode(), ]
            },
        ]
    }
    return 200, [], json.dumps(keys)


def build_adfs_meta(request):
    with open(os.path.join(os.path.dirname(__file__), "mock_files/FederationMetadata.xml"), mode="r") as f:
        data = "".join(f.readlines())
    data = data.replace("REPLACE_WITH_CERT_A", base64.b64encode(signing_cert_A).decode())
    data = data.replace("REPLACE_WITH_CERT_B", base64.b64encode(signing_cert_B).decode())
    return 200, [], data


def mock_adfs(adfs_version):
    if adfs_version not in ["2012", "2016", "azure"]:
        raise NotImplementedError("This version of ADFS is not implemented")

    def do_mock(test_func):
        def wrapper(*original_args, **original_kwargs):
            prefix_table = {
                "2012": "https://adfs.example.com",
                "2016": "https://adfs.example.com",
                "azure": "https://login.microsoftonline.com",
            }
            prefix = prefix_table[adfs_version]
            openid_cfg = re.compile(prefix + r".*\.well-known/openid-configuration")
            openid_keys = re.compile(prefix + r".*/discovery/keys")
            adfs_meta = re.compile(prefix + r".*/FederationMetadata/2007-06/FederationMetadata\.xml")
            token_endpoint = re.compile(prefix + r".*/oauth2/token")
            with responses.RequestsMock(assert_all_requests_are_fired=False) as rsps:
                # https://github.com/getsentry/responses
                if adfs_version == "2016":
                    rsps.add(
                        rsps.GET, openid_cfg,
                        json=load_json("mock_files/adfs-openid-configuration.json")
                    )
                    rsps.add_callback(
                        rsps.GET, openid_keys,
                        callback=build_openid_keys,
                        content_type='application/json',
                    )
                elif adfs_version == "azure":
                    rsps.add(
                        rsps.GET, openid_cfg,
                        json=load_json("mock_files/azure-openid-configuration.json")
                    )
                    rsps.add_callback(
                        rsps.GET, openid_keys,
                        callback=build_openid_keys,
                        content_type='application/json',
                    )
                else:
                    rsps.add(
                        rsps.GET, openid_cfg,
                        status=404
                    )
                    rsps.add(
                        rsps.GET, openid_keys,
                        status=404
                    )

                rsps.add_callback(
                    rsps.GET, adfs_meta,
                    callback=build_adfs_meta,
                    content_type='application/xml',
                )
                if adfs_version == "azure":
                    rsps.add_callback(
                        rsps.POST, token_endpoint,
                        callback=build_access_token_azure,
                        content_type='application/json',
                    )
                else:
                    rsps.add_callback(
                        rsps.POST, token_endpoint,
                        callback=build_access_token_adfs,
                        content_type='application/json',
                    )

                test_func(*original_args, **original_kwargs)
        return wrapper
    return do_mock


signing_key_A, signing_cert_A = generate_key_and_cert()
signing_key_B, signing_cert_B = generate_key_and_cert()



from copy import deepcopy

from django.core.exceptions import ImproperlyConfigured
from django.test import TestCase
from mock import patch
from django_auth_adfs.config import django_settings
from django_auth_adfs.config import Settings


class SettingsTests(TestCase):
    def test_no_settings(self):
        settings = deepcopy(django_settings)
        del settings.AUTH_ADFS
        with patch("django_auth_adfs.config.django_settings", settings):
            self.assertRaises(ImproperlyConfigured, Settings)

    def test_claim_mapping_overlapping_username_field(self):
        settings = deepcopy(django_settings)
        settings.AUTH_ADFS["CLAIM_MAPPING"] = {"username": "samaccountname"}
        with patch("django_auth_adfs.config.django_settings", settings):
            self.assertRaises(ImproperlyConfigured, Settings)

    def test_tenant_and_server(self):
        settings = deepcopy(django_settings)
        settings.AUTH_ADFS["TENEANT_ID"] = "abc"
        settings.AUTH_ADFS["server"] = "abc"
        with patch("django_auth_adfs.config.django_settings", settings):
            self.assertRaises(ImproperlyConfigured, Settings)

    def test_unknown_setting(self):
        settings = deepcopy(django_settings)
        settings.AUTH_ADFS["dummy"] = "abc"
        with patch("django_auth_adfs.config.django_settings", settings):
            self.assertRaises(ImproperlyConfigured, Settings)

    def test_required_setting(self):
        settings = deepcopy(django_settings)
        del settings.AUTH_ADFS["AUDIENCE"]
        with patch("django_auth_adfs.config.django_settings", settings):
            self.assertRaises(ImproperlyConfigured, Settings)



import base64
import logging
import warnings
from datetime import datetime, timedelta
from xml.etree import ElementTree

import requests
import requests.adapters
from cryptography.hazmat.backends.openssl.backend import backend
from cryptography.x509 import load_der_x509_certificate
from django.conf import settings as django_settings
from django.contrib.auth import REDIRECT_FIELD_NAME
from django.contrib.auth import get_user_model
from django.core.exceptions import ImproperlyConfigured
from django.http import QueryDict

try:
    from django.urls import reverse
except ImportError:  # Django < 1.10
    from django.core.urlresolvers import reverse

logger = logging.getLogger("django_auth_adfs")

AZURE_AD_SERVER = "login.microsoftonline.com"


class ConfigLoadError(Exception):
    pass


class Settings(object):
    def __init__(self):
        # Set defaults
        self.AUDIENCE = None  # Required
        self.BOOLEAN_CLAIM_MAPPING = {}
        self.CA_BUNDLE = True
        self.CLAIM_MAPPING = {}
        self.CLIENT_ID = None  # Required
        self.CLIENT_SECRET = None
        self.CONFIG_RELOAD_INTERVAL = 24  # hours
        self.DISABLE_SSO = False
        self.GROUP_TO_FLAG_MAPPING = {}
        self.GROUPS_CLAIM = "group"
        self.LOGIN_EXEMPT_URLS = []
        self.MIRROR_GROUPS = False
        self.RELYING_PARTY_ID = None  # Required
        self.RETRIES = 3
        self.SERVER = None  # Required
        self.TENANT_ID = None  # Required
        self.TIMEOUT = 5
        self.USERNAME_CLAIM = "winaccountname"

        required_settings = [
            "AUDIENCE",
            "CLIENT_ID",
            "RELYING_PARTY_ID",
            "USERNAME_CLAIM",
        ]

        deprecated_settings = {
            "AUTHORIZE_PATH": "This setting is automatically loaded from ADFS.",
            "ISSUER": "This setting is automatically loaded from ADFS.",
            "LOGIN_REDIRECT_URL": "Instead use the standard Django settings with the same name.",
            "REDIR_URI": "This setting is automatically determined based on the URL configuration of Django.",
            "SIGNING_CERT": "The token signing certificates are automatically loaded from ADFS.",
            "TOKEN_PATH": "This setting is automatically loaded from ADFS.",
        }

        if not hasattr(django_settings, "AUTH_ADFS"):
            msg = "The configuration directive 'AUTH_ADFS' was not found in your Django settings"
            raise ImproperlyConfigured(msg)

        # Handle deprecated settings
        for setting, message in deprecated_settings.items():
            if setting in django_settings.AUTH_ADFS:
                warnings.warn("Setting {} is deprecated and it's value was ignored. {}".format(setting, message),
                              DeprecationWarning)
                del django_settings.AUTH_ADFS[setting]

        if "CERT_MAX_AGE" in django_settings.AUTH_ADFS:
            django_settings.AUTH_ADFS["CONFIG_RELOAD_INTERVAL"] = django_settings.AUTH_ADFS["CERT_MAX_AGE"]
            warnings.warn('Setting CERT_MAX_AGE has been renamed to CONFIG_RELOAD_INTERVAL. The value was copied.',
                          DeprecationWarning)
            del django_settings.AUTH_ADFS["CERT_MAX_AGE"]

        if "GROUP_CLAIM" in django_settings.AUTH_ADFS:
            django_settings.AUTH_ADFS["GROUPS_CLAIM"] = django_settings.AUTH_ADFS["GROUP_CLAIM"]
            warnings.warn('Setting GROUP_CLAIM has been renamed to GROUPS_CLAIM. The value was copied.',
                          DeprecationWarning)
            del django_settings.AUTH_ADFS["GROUP_CLAIM"]

        if "RESOURCE" in django_settings.AUTH_ADFS:
            django_settings.AUTH_ADFS["RELYING_PARTY_ID"] = django_settings.AUTH_ADFS["RESOURCE"]
            del django_settings.AUTH_ADFS["RESOURCE"]

        if "TENANT_ID" in django_settings.AUTH_ADFS:
            # If a tenant ID was set, switch to Azure AD mode
            if "SERVER" in django_settings.AUTH_ADFS:
                raise ImproperlyConfigured("The SERVER cannot be set when TENANT_ID is set.")
            self.SERVER = AZURE_AD_SERVER
            self.USERNAME_CLAIM = "upn"
            self.GROUPS_CLAIM = "groups"
            self.CLAIM_MAPPING = {"first_name": "given_name",
                                  "last_name": "family_name",
                                  "email": "email"}

        # Overwrite defaults with user settings
        for setting, value in django_settings.AUTH_ADFS.items():
            if hasattr(self, setting):
                setattr(self, setting, value)
            else:
                msg = "'{0}' is not a valid configuration directive for django_auth_adfs."
                raise ImproperlyConfigured(msg.format(setting))

        # Validate required settings
        if not self.TENANT_ID and not self.SERVER:
            msg = "Exactly one of the settings TENANT_ID or SERVER must be set"
            raise ImproperlyConfigured(msg)
        elif self.TENANT_ID is None:
            # For on premises ADFS, the tenant ID is set to adfs
            # On AzureAD the adfs part in the URL happens to be replace by the tenant ID.
            self.TENANT_ID = "adfs"

        for setting in required_settings:
            if not getattr(self, setting):
                msg = "django_auth_adfs setting '{0}' has not been set".format(setting)
                raise ImproperlyConfigured(msg)

        # Validate setting conflicts
        usermodel = get_user_model()
        if usermodel.USERNAME_FIELD in self.CLAIM_MAPPING.keys():
            raise ImproperlyConfigured("You cannot set the username field of the user model from "
                                       "the CLAIM_MAPPING setting. Instead use the USERNAME_CLAIM setting.")


class ProviderConfig(object):
    def __init__(self):
        self._config_timestamp = None
        self._mode = None

        self.authorization_endpoint = None
        self.signing_keys = None
        self.token_endpoint = None
        self.end_session_endpoint = None
        self.issuer = None

        self.session = requests.Session()
        adapter = requests.adapters.HTTPAdapter(max_retries=settings.RETRIES)
        self.session.mount('https://', adapter)
        self.session.verify = settings.CA_BUNDLE

    def load_config(self):
        # If loaded data is too old, reload it again
        refresh_time = datetime.now() - timedelta(hours=settings.CONFIG_RELOAD_INTERVAL)
        if self._config_timestamp is None or self._config_timestamp < refresh_time:
            logger.debug("Loading django_auth_adfs ID Provider configuration.")
            try:
                loaded = self._load_openid_config()
                self._mode = "openid_connect"
            except ConfigLoadError:
                loaded = self._load_federation_metadata()
                self._mode = "oauth2"

            if not loaded:
                if self._config_timestamp is None:
                    msg = "Could not load any data from ADFS server. "\
                          "Authentication against ADFS not be possible. "\
                          "Verify your settings and the connection with the ADFS server."
                    logger.critical(msg)
                    raise RuntimeError(msg)
                else:
                    # We got data from the previous time. Log a message, but don't abort.
                    logger.warning("Could not load any data from ADFS server. Keeping previous configurations")
            self._config_timestamp = datetime.now()

            logger.info("django_auth_adfs loaded settings from ADFS server.")
            logger.info("operating mode:         " + self._mode)
            logger.info("authorization endpoint: " + self.authorization_endpoint)
            logger.info("token endpoint:         " + self.token_endpoint)
            logger.info("end session endpoint:   " + self.end_session_endpoint)
            logger.info("issuer:                 " + self.issuer)

    def _load_openid_config(self):
        config_url = "https://{}/{}/.well-known/openid-configuration".format(
            settings.SERVER, settings.TENANT_ID
        )

        try:
            logger.info("Trying to get OpenID Connect config from {}".format(config_url))
            response = self.session.get(config_url, timeout=settings.TIMEOUT)
            response.raise_for_status()
            openid_cfg = response.json()

            response = self.session.get(openid_cfg["jwks_uri"], timeout=settings.TIMEOUT)
            response.raise_for_status()
            signing_certificates = [x["x5c"][0] for x in response.json()["keys"] if x.get("use", "sig") == "sig"]
            #                               ^^^
            # https://tools.ietf.org/html/draft-ietf-jose-json-web-key-41#section-4.7
            # The PKIX certificate containing the key value MUST be the first certificate
        except requests.HTTPError:
            raise ConfigLoadError

        self._load_keys(signing_certificates)
        try:
            self.authorization_endpoint = openid_cfg["authorization_endpoint"]
            self.token_endpoint = openid_cfg["token_endpoint"]
            self.end_session_endpoint = openid_cfg["end_session_endpoint"]
            if settings.TENANT_ID is not 'adfs':
                self.issuer = openid_cfg["issuer"]
            else:
                self.issuer = openid_cfg["access_token_issuer"]
        except KeyError:
            raise ConfigLoadError
        return True

    def _load_federation_metadata(self):
        server_url = "https://{}".format(settings.SERVER)
        base_url = "{}/{}".format(server_url, settings.TENANT_ID)
        if settings.TENANT_ID == "adfs":
            adfs_config_url = server_url + "/FederationMetadata/2007-06/FederationMetadata.xml"
        else:
            adfs_config_url = base_url + "/FederationMetadata/2007-06/FederationMetadata.xml"

        try:
            logger.info("Trying to get ADFS Metadata file {}".format(adfs_config_url))
            response = self.session.get(adfs_config_url, timeout=settings.TIMEOUT)
            response.raise_for_status()
        except requests.HTTPError:
            raise ConfigLoadError

        # Extract token signing certificates
        xml_tree = ElementTree.fromstring(response.content)
        cert_nodes = xml_tree.findall(
            "./{urn:oasis:names:tc:SAML:2.0:metadata}RoleDescriptor"
            "[@{http://www.w3.org/2001/XMLSchema-instance}type='fed:SecurityTokenServiceType']"
            "/{urn:oasis:names:tc:SAML:2.0:metadata}KeyDescriptor[@use='signing']"
            "/{http://www.w3.org/2000/09/xmldsig#}KeyInfo"
            "/{http://www.w3.org/2000/09/xmldsig#}X509Data"
            "/{http://www.w3.org/2000/09/xmldsig#}X509Certificate")
        signing_certificates = [node.text for node in cert_nodes]

        self._load_keys(signing_certificates)
        self.issuer = xml_tree.get("entityID")
        self.authorization_endpoint = base_url + "/oauth2/authorize"
        self.token_endpoint = base_url + "/oauth2/token"
        self.end_session_endpoint = base_url + "/ls/?wa=wsignout1.0"
        return True

    def _load_keys(self, certificates):
        new_keys = []
        for cert in certificates:
            logger.debug("Loading public key from certificate: " + cert)
            cert_obj = load_der_x509_certificate(base64.b64decode(cert), backend)
            new_keys.append(cert_obj.public_key())
        self.signing_keys = new_keys

    def redirect_uri(self, request):
        self.load_config()
        return request.build_absolute_uri(reverse("django_auth_adfs:callback"))

    def build_authorization_endpoint(self, request, disable_sso=None):
        """
        This function returns the ADFS authorization URL.

        Args:
            request(django.http.request.HttpRequest): A django Request object
            disable_sso(bool): Whether to disable single sign-on and force the ADFS server to show a login prompt.

        Returns:
            str: The redirect URI

        """
        self.load_config()
        redirect_to = request.GET.get(REDIRECT_FIELD_NAME, None)
        if not redirect_to:
            redirect_to = django_settings.LOGIN_REDIRECT_URL
        redirect_to = base64.urlsafe_b64encode(redirect_to.encode()).decode()
        query = QueryDict(mutable=True)
        query.update({
            "response_type": "code",
            "client_id": settings.CLIENT_ID,
            "resource": settings.RELYING_PARTY_ID,
            "redirect_uri": self.redirect_uri(request),
            "state": redirect_to,
        })
        if self._mode == "openid_connect":
            query["scope"] = "openid"
            if (disable_sso is None and settings.DISABLE_SSO) or disable_sso is True:
                query["prompt"] = "login"

        return "{0}?{1}".format(self.authorization_endpoint, query.urlencode())

    def build_end_session_endpoint(self):
        """
        This function returns the ADFS end session URL to log a user out.

        Returns:
            str: The redirect URI

        """
        self.load_config()
        return self.end_session_endpoint


settings = Settings()
provider_config = ProviderConfig()



SECRET_KEY = 'secret'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
        'USER': '',
        'PASSWORD': '',
        'HOST': '',
        'PORT': '',
    }
}

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
        'DIRS': 'templates'
    },
]

MIDDLEWARE = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',

    'django_auth_adfs.middleware.LoginRequiredMiddleware',
)
# Django < 1.10 compatibility
MIDDLEWARE_CLASSES = MIDDLEWARE

INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',

    'django_auth_adfs',
    'tests',
)

AUTHENTICATION_BACKENDS = (
    "django.contrib.auth.backends.ModelBackend",
    'django_auth_adfs.backend.AdfsAuthCodeBackend',
    'django_auth_adfs.backend.AdfsAccessTokenBackend',
)

ROOT_URLCONF = 'tests.urls'

STATIC_ROOT = '/tmp/'  # Dummy
STATIC_URL = '/static/'

AUTH_ADFS = {
    "SERVER": "adfs.example.com",
    "CLIENT_ID": "your-configured-client-id",
    "RELYING_PARTY_ID": "your-adfs-RPT-name",
    "AUDIENCE": "microsoft:identityserver:your-RelyingPartyTrust-identifier",
    "CA_BUNDLE": "/path/to/ca-bundle.pem",
    "CLAIM_MAPPING": {"first_name": "given_name",
                      "last_name": "family_name",
                      "email": "email"},
    "BOOLEAN_CLAIM_MAPPING": {"is_staff": "user_is_staff",
                              "is_superuser": "user_is_superuser"},
    "LOGIN_EXEMPT_URLS": ["^context_processor/$"]
}



from copy import deepcopy

from django.contrib.auth.models import User, Group
from django.core.exceptions import PermissionDenied
from django.test import TestCase, RequestFactory
from mock import Mock, patch

from django_auth_adfs import signals
from django_auth_adfs.backend import AdfsAuthCodeBackend
from django_auth_adfs.config import ProviderConfig, Settings
from .utils import mock_adfs


class AuthenticationTests(TestCase):
    def setUp(self):
        Group.objects.create(name='group1')
        Group.objects.create(name='group2')
        Group.objects.create(name='group3')
        self.request = RequestFactory().get('/oauth2/callback')
        self.signal_handler = Mock()
        signals.post_authenticate.connect(self.signal_handler)

    @mock_adfs("2012")
    def test_post_authenticate_signal_send(self):
        backend = AdfsAuthCodeBackend()
        backend.authenticate(self.request, authorization_code="dummycode")
        self.assertEqual(self.signal_handler.call_count, 1)

    @mock_adfs("2012")
    def test_with_auth_code_2012(self):
        backend = AdfsAuthCodeBackend()
        user = backend.authenticate(self.request, authorization_code="dummycode")
        self.assertIsInstance(user, User)
        self.assertEqual(user.first_name, "John")
        self.assertEqual(user.last_name, "Doe")
        self.assertEqual(user.email, "john.doe@example.com")
        self.assertEqual(len(user.groups.all()), 2)
        self.assertEqual(user.groups.all()[0].name, "group1")
        self.assertEqual(user.groups.all()[1].name, "group2")

    @mock_adfs("2016")
    def test_with_auth_code_2016(self):
        backend = AdfsAuthCodeBackend()
        user = backend.authenticate(self.request, authorization_code="dummycode")
        self.assertIsInstance(user, User)
        self.assertEqual(user.first_name, "John")
        self.assertEqual(user.last_name, "Doe")
        self.assertEqual(user.email, "john.doe@example.com")
        self.assertEqual(len(user.groups.all()), 2)
        self.assertEqual(user.groups.all()[0].name, "group1")
        self.assertEqual(user.groups.all()[1].name, "group2")

    @mock_adfs("azure")
    def test_with_auth_code_azure(self):
        from django_auth_adfs.config import django_settings
        settings = deepcopy(django_settings)
        del settings.AUTH_ADFS["SERVER"]
        settings.AUTH_ADFS["TENANT_ID"] = "dummy_tenant_id"
        with patch("django_auth_adfs.config.django_settings", settings):
            with patch("django_auth_adfs.config.settings", Settings()):
                with patch("django_auth_adfs.backend.provider_config", ProviderConfig()):
                    backend = AdfsAuthCodeBackend()
                    user = backend.authenticate(self.request, authorization_code="dummycode")
                    self.assertIsInstance(user, User)
                    self.assertEqual(user.first_name, "John")
                    self.assertEqual(user.last_name, "Doe")
                    self.assertEqual(user.email, "john.doe@example.com")
                    self.assertEqual(len(user.groups.all()), 2)
                    self.assertEqual(user.groups.all()[0].name, "group1")
                    self.assertEqual(user.groups.all()[1].name, "group2")

    @mock_adfs("2016")
    def test_empty(self):
        backend = AdfsAuthCodeBackend()
        self.assertIsNone(backend.authenticate(self.request))

    @mock_adfs("2016")
    def test_group_claim(self):
        backend = AdfsAuthCodeBackend()
        with patch("django_auth_adfs.backend.settings.GROUPS_CLAIM", "nonexisting"):
            user = backend.authenticate(self.request, authorization_code="dummycode")
            self.assertIsInstance(user, User)
            self.assertEqual(user.first_name, "John")
            self.assertEqual(user.last_name, "Doe")
            self.assertEqual(user.email, "john.doe@example.com")
            self.assertEqual(len(user.groups.all()), 0)

    @mock_adfs("2016")
    def test_empty_keys(self):
        backend = AdfsAuthCodeBackend()
        with patch("django_auth_adfs.config.provider_config.signing_keys", []):
            self.assertRaises(PermissionDenied, backend.authenticate, self.request, authorization_code='testcode')

    @mock_adfs("2016")
    def test_group_removal(self):
        user, created = User.objects.get_or_create(**{
            User.USERNAME_FIELD: "testuser"
        })
        group = Group.objects.get(name="group3")
        user.groups.add(group)
        user.set_unusable_password()
        user.save()

        self.assertEqual(user.groups.all()[0].name, "group3")

        backend = AdfsAuthCodeBackend()

        user = backend.authenticate(self.request, authorization_code="dummycode")
        self.assertIsInstance(user, User)
        self.assertEqual(user.first_name, "John")
        self.assertEqual(user.last_name, "Doe")
        self.assertEqual(user.email, "john.doe@example.com")
        self.assertEqual(len(user.groups.all()), 2)
        self.assertEqual(user.groups.all()[0].name, "group1")
        self.assertEqual(user.groups.all()[1].name, "group2")



import json
from copy import deepcopy

from django.test import RequestFactory, TestCase

from django_auth_adfs.config import ProviderConfig, Settings
from django_auth_adfs.rest_framework import AdfsAccessTokenAuthentication

from mock import patch

from rest_framework import exceptions


from .utils import build_access_token_adfs, build_access_token_azure, mock_adfs


class RestFrameworkIntegrationTests(TestCase):
    def setUp(self):
        self.drf_auth_class = AdfsAccessTokenAuthentication()

        adfs_response = build_access_token_adfs(RequestFactory().get('/'))[2]
        self.access_token_adfs = json.loads(adfs_response)['access_token']

        azure_response = build_access_token_azure(RequestFactory().get('/'))[2]
        self.access_token_azure = json.loads(azure_response)['access_token']

    @mock_adfs("2012")
    def test_access_token_2012(self):
        access_token_header = "Bearer {}".format(self.access_token_adfs)
        request = RequestFactory().get('/api', HTTP_AUTHORIZATION=access_token_header)

        user, token = self.drf_auth_class.authenticate(request)
        self.assertEqual(user.username, "testuser")
        self.assertEqual(token, self.access_token_adfs.encode())

    @mock_adfs("2016")
    def test_access_token_2016(self):
        access_token_header = "Bearer {}".format(self.access_token_adfs)
        request = RequestFactory().get('/api', HTTP_AUTHORIZATION=access_token_header)

        user, token = self.drf_auth_class.authenticate(request)
        self.assertEqual(user.username, "testuser")
        self.assertEqual(token, self.access_token_adfs.encode())

    @mock_adfs("azure")
    def test_access_token_azure(self):
        access_token_header = "Bearer {}".format(self.access_token_azure)
        request = RequestFactory().get('/api', HTTP_AUTHORIZATION=access_token_header)

        from django_auth_adfs.config import django_settings
        settings = deepcopy(django_settings)
        del settings.AUTH_ADFS["SERVER"]
        settings.AUTH_ADFS["TENANT_ID"] = "dummy_tenant_id"
        with patch("django_auth_adfs.config.django_settings", settings):
            with patch("django_auth_adfs.config.settings", Settings()):
                with patch("django_auth_adfs.backend.provider_config", ProviderConfig()):
                    user, token = self.drf_auth_class.authenticate(request)
                    self.assertEqual(user.username, "testuser")

    @mock_adfs("2012")
    def test_access_token_exceptions(self):
        access_token_header = "Bearer non-existing-token"
        request = RequestFactory().get('/api', HTTP_AUTHORIZATION=access_token_header)

        with self.assertRaises(exceptions.AuthenticationFailed):
            self.drf_auth_class.authenticate(request)

        # use the azure token on adfs should not work
        access_token_header = "Bearer {}".format(self.access_token_azure)
        request = RequestFactory().get('/api', HTTP_AUTHORIZATION=access_token_header)

        with self.assertRaises(exceptions.AuthenticationFailed):
            self.drf_auth_class.authenticate(request)
